#ifndef __DISKDEV_H
#define __DISKDEV_H


#include "systypes.h"

#define BYTES_PERSECTOR 512
#define MAX_HD 4
#define MAX_LDEV 16
#define TYPE_EXT_PARTITION 5

#define BLOCKSIZE(d)	(d->sec_perblock*512)

typedef struct 
{
	uint8 		boot_ind;		/* 0x80 - active */
	uint8 		head;			/* starting head */
	uint8 		sector;			/* starting sector */
	uint8 		cyl;			/* starting cylinder */
	uint8 		sys_ind;		/* What partition type */
	uint8 		end_head;		/* end head */
	uint8 		end_sector;		/* end sector */
	uint8 		end_cyl;		/* end cylinder */
	uint32 		start_sec;		/* starting sector counting from 0 */
	uint32 		nr_sec;			/* nr of sectors in partition */
} partition_t;


typedef struct 
{
	uint32		maxhead;
	uint32		maxcyl;
	uint32		maxsec;
} pdev_t;

int	pdev_read(uint8 disk, uint32 start_sec, uint8 nr_sec, void far *data);
int	pdev_write(uint8 disk, uint32 start_sec, uint8 nr_sec, void far * data);

typedef struct 
{
	int32 		start_sec;
	int32 		nr_sec;

	uint8		disk;					/* disk number				*/
	uint8 		type;					/* match sys_ind above		*/
	uint8		partition;				/* partion number			*/
	uint8		sec_perblock;			/* # sectors per block		*/
	void		*fs_info;				/* file system depend info	*/
} ldev_t;

int		ldev_read(ldev_t *ldev, uint32 start_sec, uint8 nr_sec, void far *data);
int		ldev_write(ldev_t *ldev, uint32 start_sec, uint8 nr_sec, void far *data);
int		ldev_readblock(ldev_t *ldev, uint32 block, void far *data);
int		ldev_writeblock(ldev_t *ldev, uint32 block, void far *data);
int		ldev_setblocksize(ldev_t *ldev, uint32 bsize);
void 	pdev_readpartitions();


#ifdef DECL_DEVICES
pdev_t pDevices[MAX_HD];
ldev_t lDevices[MAX_HD][MAX_LDEV];
#else
extern pdev_t pDevices[MAX_HD];
extern ldev_t lDevices[MAX_HD][MAX_LDEV];
#endif
#undef DECL_DEVICES


#endif		// __DISKDEV_H
