#pragma check_stack(off)
#include <stdlib.h>
#include <string.h>
#include <dos.h>

#ifdef __BORLANDC__
#include <alloc.h>
#include <mem.h>
#else
#include <malloc.h>
#endif

//#include "ll_mem.h"
#include "debug.h"
#include "diskdev.h"
#include "ext2_fs.h"
#include "emscache.h"
#include "tsr.h"


#define	CACHE_PAGE		3
emshandle_t				emsCachePage=0;
unsigned				MapSP=(unsigned) CACHE_PAGE* (unsigned) PAGE_SIZE;
char far*				CurCachePointer;

buffer_cache_t			*InodeCache;	/* Inode cache */
buffer_cache_t			*BlockCache;	/* Block cache */

/*
 * some neat shortcuts
 */
#define INODE_BMAP(inode,nr) ((inode)->i_block[(nr)])
#define BLOCK_BMAP(data,nr) (((unsigned long far *) data)[(nr)])
#define SUPERBLOCK(d)	(((ext2_fs_info_t *)d->fs_info)->ext2_sb)
#define SBINFO(d)		(((ext2_fs_info_t *)d->fs_info)->ext2_sbinfo)

/*
 * In order for other routines to read the contents of inodes and
 * diskblocks (which both reside in the emscache), we need to map
 * these items in conventional memory. For this we use the ems
 * PageFrame. Because we need some items at the same time, we use
 * a stack mechanism to push and pop items in the PageFrame. The
 * only thing these Push and Pop routines do is adjusting the Stack
 * Pointer.
 */

static char far* PushItem(unsigned size)
{
	char far *rval=CurCachePointer;

	if (MapSP + size < MapSP)
	{
				/* should not happen */
		press_key("PushItem wrapped around MapSP! Press a key to continue, but better is to reboot\n");
	}
	MapSP+=size;
	CurCachePointer+=size;
	return rval;
}

static void PopItem(unsigned size)
{
	if (MapSP - size < (unsigned) CACHE_PAGE* (unsigned) PAGE_SIZE)
	{
				/* should not happen */
		press_key("PopItem wrapped around MapSP! Press a key to continue, but better is to reboot\n");
	}
	MapSP-=size;
	CurCachePointer-=size;
}


static int ext2_read_sb(ldev_t *ldev)
{
	char far *			buffer;
	ext2_super_block 	*sb=SUPERBLOCK(ldev);
	ext2_sb_info		*sbinfo=SBINFO(ldev);
	uint32 				logic_sb_block;
	uint32 				desc_size;
	int 				db_count,i;

/*
 * The superblock is the logicalblock 1 on disk and has a size
 * of 1024 bytes
 */

	buffer=PushItem(EXT2_SIZE_SB);
	ldev_setblocksize(ldev,EXT2_SIZE_SB);
	if (!ldev_readblock(ldev,1,(void far *) sb))
	{
		dprintf(D_EXT2,"Could not read block for superblock\n");
		PopItem(EXT2_SIZE_SB);
		return 0;
	}
	//_fmemcpy(sb, buffer, sizeof(ext2_super_block));
	PopItem(EXT2_SIZE_SB);

	if (sb->s_magic != EXT2_SUPER_MAGIC)
	{
		dprintf(D_EXT2,"Magic check failed\n");
		return 0;
	}
/*
 * Set the blocksize for this device
 */
	if (!ldev_setblocksize(ldev,EXT2_BLOCK_SIZE(sb)))
	{
		dprintf(D_EXT2,"Could not set blocksize:%u\n",(unsigned) EXT2_BLOCK_SIZE(sb));
		return 0;
	}

	logic_sb_block=sb->s_first_data_block;
	sbinfo->s_frag_size = EXT2_MIN_FRAG_SIZE <<  sb->s_log_frag_size;
	if (sbinfo->s_frag_size)
		sbinfo->s_frags_per_block = EXT2_BLOCK_SIZE(sb) / sbinfo->s_frag_size;
	else
		sb->s_magic = 0;
	sbinfo->s_blocks_per_group = sb->s_blocks_per_group;
	sbinfo->s_frags_per_group = sb->s_frags_per_group;
	sbinfo->s_inodes_per_group = sb->s_inodes_per_group;
	sbinfo->s_inodes_per_block = EXT2_BLOCK_SIZE(sb) / sizeof (ext2_inode);
	sbinfo->s_itb_per_group = sbinfo->s_inodes_per_group / sbinfo->s_inodes_per_block;
	sbinfo->s_desc_per_block = EXT2_BLOCK_SIZE(sb) / sizeof (ext2_group_desc);


	sbinfo->s_groups_count = (sb->s_blocks_count -sb->s_first_data_block +
					   EXT2_BLOCKS_PER_GROUP(sb) - 1) /
				       EXT2_BLOCKS_PER_GROUP(sb);
	db_count = (sbinfo->s_groups_count + EXT2_DESC_PER_BLOCK(sbinfo) - 1) /
		   EXT2_DESC_PER_BLOCK(sbinfo);
	sbinfo->s_group_desc = (ext2_group_desc **) tsr_malloc (db_count * sizeof (ext2_group_desc *));
	if (sbinfo->s_group_desc == NULL)
	{
		dprintf(D_EXT2,"EXT2-fs: not enough memory\n");
		return 0;
	}
	for (i = 0; i < db_count; i++)
	{
		buffer=PushItem(BLOCKSIZE(ldev));
		if (!ldev_readblock(ldev,logic_sb_block + i + 1,buffer))
		{
			dprintf(D_EXT2,"Could not read block for group desc\n");
			PopItem(BLOCKSIZE(ldev));
			return 0;
		}
		desc_size=sizeof(ext2_group_desc)*EXT2_DESC_PER_BLOCK(sbinfo);
		if ( !(sbinfo->s_group_desc[i]=(ext2_group_desc *)tsr_malloc((size_t) desc_size)))
		{
			dprintf(D_EXT2,"Could not alloc group desc %i\n",i);
			PopItem(BLOCKSIZE(ldev));
			return 0;
		}
		//far_2_default(sbinfo->s_group_desc[i],buffer,(unsigned) desc_size);
		_fmemcpy(sbinfo->s_group_desc[i],buffer,(unsigned) desc_size);
		PopItem(BLOCKSIZE(ldev));
	}
/*
 * We should check the group descriptors here but we leave that
 * for the future :-)
 */
	for (i = 0; i < EXT2_MAX_GROUP_LOADED; i++)
	{
		sbinfo->s_inode_bitmap_number[i] = 0;
		sbinfo->s_inode_bitmap[i] = NULL;
		sbinfo->s_block_bitmap_number[i] = 0;
		sbinfo->s_block_bitmap[i] = NULL;
	}
	sbinfo->s_loaded_inode_bitmaps = 0;
	sbinfo->s_loaded_block_bitmaps = 0;
	sbinfo->s_db_per_group = db_count;

	return 1;
}

cache_entry_t* ext2_readinode(ldev_t *ldev, unsigned l_unit, uint32 i_ino)
{
	char far 		*buffer;
	unsigned		emsPage;
	unsigned long	offset;
	ext2_inode 		far *raw_inode;
	uint32 			block_group;
	uint32 			group_desc;
	uint32 			desc;
	uint32 			block;
	ext2_group_desc *gdp;
	cache_entry_t	*entry;

	ext2_super_block	*sb=SUPERBLOCK(ldev);
	ext2_sb_info		*sbinfo=SBINFO(ldev);

	if ((i_ino != EXT2_ROOT_INO && i_ino != EXT2_ACL_IDX_INO &&
		 i_ino != EXT2_ACL_DATA_INO && i_ino < EXT2_FIRST_INO) ||
		i_ino > sb->s_inodes_count)
	{
		dprintf(D_EXT2,"ext2_read_inode,bad inode number: %lu\n", i_ino);
		return 0;
	}

	buffer=PushItem(sizeof(ext2_inode));
	if ((entry=cache_lookup(InodeCache,l_unit,i_ino)))
	{
		offset= (unsigned long) InodeCache->size_item * entry->emsIndex;
		emsPage=(unsigned) (offset / (unsigned long) PAGE_SIZE);
		offset=offset % (unsigned long) PAGE_SIZE;
		emsEms2Ems(InodeCache->emsHandle,emsPage,(unsigned)
				offset,emsCachePage,0, FP_OFF(buffer)-(CACHE_PAGE * (unsigned) PAGE_SIZE),InodeCache->size_item);
		entry->data=buffer;

		return entry;
	}

	block_group = (i_ino - 1) / EXT2_INODES_PER_GROUP(sbinfo);
	if (block_group >= sbinfo->s_groups_count)
	{
		dprintf(D_EXT2,"ext2_read_inode, group >= groups count\n");
		PopItem(sizeof(ext2_inode));
		return 0;
	}

	group_desc = block_group / EXT2_DESC_PER_BLOCK(sbinfo);
	desc = block_group % EXT2_DESC_PER_BLOCK(sbinfo);
	gdp = sbinfo->s_group_desc[group_desc];
	block = gdp[desc].bg_inode_table +	(((i_ino - 1) % EXT2_INODES_PER_GROUP(sb)) / EXT2_INODES_PER_BLOCK(sb));

	entry=ext2_readblock(ldev,l_unit,block);
	raw_inode = ((ext2_inode far *) entry->data) + (i_ino - 1) % EXT2_INODES_PER_BLOCK(sb);
	_fmemcpy(buffer,raw_inode,sizeof(ext2_inode));
	ext2_releaseblock(ldev,entry);
	entry=cache_add(InodeCache,l_unit,i_ino,buffer);
	entry->data=buffer;

	return entry;
}

void ext2_releaseinode(ldev_t *ldev, cache_entry_t *entry)
{
	PopItem(sizeof(ext2_inode));
}

uint32 ext2_bmap(ldev_t *ldev,unsigned l_unit, ext2_inode far *inode, int32 block)
{
	cache_entry_t			*entry;
	uint32 				rval;
	ext2_super_block	*sb=SUPERBLOCK(ldev);
	int32 				addr_per_block = EXT2_ADDR_PER_BLOCK(sb);

	if (block < 0)
	{
		dprintf(D_EXT2,"ext2_bmap,block < 0\n");
		return 0l;
	}

	if (block >= EXT2_NDIR_BLOCKS + addr_per_block +
			 addr_per_block * addr_per_block +
			 addr_per_block * addr_per_block * addr_per_block)
	{
		dprintf(D_EXT2,"ext2_bmap, block > big\n");
		return 0l;
	}

	if (block < EXT2_NDIR_BLOCKS)
		return (INODE_BMAP(inode, block));

	block -= EXT2_NDIR_BLOCKS;
	if (block < addr_per_block)
	{
		rval = INODE_BMAP(inode, EXT2_IND_BLOCK);
		if (!rval)
			return 0l;

		entry=ext2_readblock(ldev,l_unit,rval);
		rval=BLOCK_BMAP(entry->data,block);
		ext2_releaseblock(ldev,entry);
		return rval;
	}

	block -= addr_per_block;
	if (block < addr_per_block * addr_per_block)
	{
		rval = INODE_BMAP(inode, EXT2_DIND_BLOCK);
		if (!rval)
			return 0l;
		entry=ext2_readblock(ldev,l_unit,rval);
		rval = BLOCK_BMAP(entry->data,block / addr_per_block);
		ext2_releaseblock(ldev,entry);
		if (!rval)
			return 0l;

		entry=ext2_readblock(ldev,l_unit,rval);
		rval=BLOCK_BMAP(entry->data,block & (addr_per_block - 1));
		ext2_releaseblock(ldev,entry);
		return rval;
	}
	block -= addr_per_block * addr_per_block;
	rval = INODE_BMAP(inode, EXT2_TIND_BLOCK);
	if (!rval)
		return 0l;

	entry=ext2_readblock(ldev,l_unit,rval);
	rval = BLOCK_BMAP(entry->data,block / (addr_per_block * addr_per_block));
	ext2_releaseblock(ldev,entry);

	if (!rval)
		return 0l;

	entry=ext2_readblock(ldev,l_unit,rval);
	rval = BLOCK_BMAP(entry->data,(block / addr_per_block) & (addr_per_block - 1));
	ext2_releaseblock(ldev,entry);
	if (!rval)
		return 0l;

	entry=ext2_readblock(ldev,l_unit,rval);
	rval=BLOCK_BMAP(entry->data,block & (addr_per_block - 1));
	ext2_releaseblock(ldev,entry);

	return rval;
}

cache_entry_t* ext2_readblock(ldev_t *ldev,unsigned l_unit, uint32 lblock)
{
	cache_entry_t		*entry;
	char far *			buffer=PushItem(BLOCKSIZE(ldev));
	unsigned			emsPage;
	unsigned long		offset;


	dprintf(D_EXT2,"ext2_readblock %lu\n",lblock);
	if ((entry=cache_lookup(BlockCache,l_unit,lblock)))
	{
		offset= (unsigned long) BlockCache->size_item * entry->emsIndex;
		emsPage=(unsigned) (offset / (unsigned long) PAGE_SIZE);
		offset=offset % (unsigned long) PAGE_SIZE;

		emsEms2Ems(BlockCache->emsHandle,emsPage,(unsigned)
				offset,emsCachePage,0, FP_OFF(buffer)-(CACHE_PAGE * (unsigned) PAGE_SIZE),BLOCKSIZE(ldev));
	}
	else
	{
		ldev_readblock(ldev,lblock,buffer);
		entry=cache_add(BlockCache,l_unit,lblock,buffer);
	}
	entry->data=(char far *) buffer;
	return entry;
}

void ext2_releaseblock(ldev_t *ldev, cache_entry_t *entry)
{
	PopItem(BLOCKSIZE(ldev));
}


/*
 * Oh boy, this is realy ugly...
 * Shall we restart the discusion on the use of gotot's ?
 *
 * The order in which we initialize is important. We start by
 * allocating the emsPageCache (which is used as a 64kb buffer), 
 * allocate some memory (near heap!), read the superblock and
 * finaly allocate both inode and blockcache. In each routine
 * we can fail, in which case we must clean things up.
 */

int ext2_init(ldev_t *ldev, unsigned inodeCachePages, unsigned blockCachePages)
{

	if ( !(emsCachePage=emsAllocPages(1)))
	{
		dprintf(D_ALWAYS,"Could not alloc memory for emsCachePage\n");
		goto error_ext2_init_1;
	}

	if (!emsMapPage(emsCachePage,CACHE_PAGE,0))
	{
		dprintf(D_ALWAYS,"Could not map cache page\n");
		goto error_ext2_init_2;
	}
	CurCachePointer=emsPageFrame+(unsigned) CACHE_PAGE*(unsigned) PAGE_SIZE;

	if (!(ldev->fs_info=(ext2_fs_info_t *) tsr_calloc(1,sizeof(ext2_fs_info_t))))
	{
		dprintf(D_ALWAYS,"Could not alloc memory for ext2_file_info\n");
		goto error_ext2_init_2;
	}

	if (!(SUPERBLOCK(ldev)=(ext2_super_block *) tsr_calloc(1,sizeof(ext2_super_block))))
	{
		dprintf(D_ALWAYS,"Could not alloc memory for ext2_super_block\n");
		goto error_ext2_init_3;
	}

	if (!(SBINFO(ldev)=(ext2_sb_info *) tsr_calloc(1,sizeof(ext2_sb_info))))
	{
		dprintf(D_ALWAYS,"Could not alloc memory for sbinfo\n");
		goto error_ext2_init_4;
	}

	if (!ext2_read_sb(ldev))
	{
		dprintf(D_ALWAYS,"Could not read super block\n");
		goto error_ext2_init_5;
	}

	dprintf(D_SYSTEM,"Making inodecache...\n");
	if (!(InodeCache=mk_cache(inodeCachePages,sizeof(ext2_inode))))
	{
		dprintf(D_ALWAYS,"Could not make inode cache\n");
		goto error_ext2_init_5;
	}

	dprintf(D_SYSTEM,"Making blockcache...\n");
	if (!(BlockCache=mk_cache(blockCachePages,BLOCKSIZE(ldev))))
	{
		dprintf(D_ALWAYS,"Could not make block cache\n");
		goto error_ext2_init_6;
	}
	return 1;

error_ext2_init_6:
	dl_cache(InodeCache);

error_ext2_init_5:
	free(SBINFO(ldev));

error_ext2_init_4:
	free(SUPERBLOCK(ldev));

error_ext2_init_3:
	free(ldev->fs_info);

error_ext2_init_2:
	emsFreePages(emsCachePage);
	emsCachePage=0;


error_ext2_init_1:

	return 0;

}

/*
 * Free the caches, and the emsCachePage, and the malloc'ed memory.
 * NB restore the context first, otherwise emsCachePage wouldn't be freed
 */

void ext2_shutdown(ldev_t *ldev)
{
	dprintf(D_SYSTEM,"Deleting blockcache...\n");
	dl_cache(BlockCache);
	dprintf(D_SYSTEM,"Deleting inodecache...\n");
	dl_cache(InodeCache);
	emsRestoreMap(emsCachePage);
	emsFreePages(emsCachePage);
	emsCachePage=0;
#ifndef RUNNING_AS_TSR
	free(SBINFO(ldev));
	free(SUPERBLOCK(ldev));
	free(ldev->fs_info);
#endif
}

long ext2_blockscount(ldev_t *ldev)
{
	return SUPERBLOCK(ldev)->s_blocks_count;
}

long ext2_freeblockscount(ldev_t *ldev)
{
	return SUPERBLOCK(ldev)->s_free_blocks_count;
}


void ext2_savecontext(ldev_t *ldev)
{
	emsSaveMap(emsCachePage);
	emsMapPage(emsCachePage,CACHE_PAGE,0);
}

void ext2_restorecontext(ldev_t *ldev)
{
	if (emsCachePage)
		emsRestoreMap(emsCachePage);
}
