#ifndef __EXT2_FS_H
#define __EXT2_FS_H

#include "systypes.h"
#include "emscache.h"
#include "diskdev.h"


/*
 * Most of these typedefs have shamelessly been copied from the
 * linux source
 */


#define EXT2_SUPER_MAGIC	0xEF53
#define EXT2_SIZE_SB		1024			/* size of superblock on disk */
/*
 * Constants relative to the data blocks
 */
#define	EXT2_NDIR_BLOCKS		12l
#define	EXT2_IND_BLOCK			EXT2_NDIR_BLOCKS
#define	EXT2_DIND_BLOCK			(EXT2_IND_BLOCK + 1l)
#define	EXT2_TIND_BLOCK			(EXT2_DIND_BLOCK + 1l)
#define	EXT2_N_BLOCKS			(EXT2_TIND_BLOCK + 1l)

/*
 * Inode flags
 */
#define	EXT2_SECRM_FL			0x00000001l /* Secure deletion */
#define	EXT2_UNRM_FL			0x00000002l /* Undelete */
#define	EXT2_COMPR_FL			0x00000004l /* Compress file */
#define EXT2_SYNC_FL			0x00000008l /* Synchronous updates */
#define EXT2_IMMUTABLE_FL		0x00000010l /* Immutable file */
#define EXT2_APPEND_FL			0x00000020l /* writes to file may only append */
#define EXT2_NODUMP_FL			0x00000040l /* do not dump file */


/*
 * Macro-instructions used to manage several block sizes
 */

#define EXT2_MIN_BLOCK_SIZE		1024l
#define	EXT2_MAX_BLOCK_SIZE		4096l
#define EXT2_MIN_BLOCK_LOG_SIZE		  10l
#define EXT2_BLOCK_SIZE(s)	( EXT2_MIN_BLOCK_SIZE << (s)->s_log_block_size)
#define	EXT2_INODES_PER_BLOCK(s)	(EXT2_BLOCK_SIZE(s) / sizeof (ext2_inode))
#define EXT2_ACLE_PER_BLOCK(s)		(EXT2_BLOCK_SIZE(s) / sizeof (ext2_acl_entry))
#define	EXT2_ADDR_PER_BLOCK(s)		(EXT2_BLOCK_SIZE(s) / sizeof (uint32))

/*
 * Macro-instructions used to manage fragments
 */
#define EXT2_MIN_FRAG_SIZE		1024l
#define	EXT2_MAX_FRAG_SIZE		4096l
#define EXT2_MIN_FRAG_LOG_SIZE		  10l


#define EXT2_FRAG_SIZE(s)			( (s)->s_frag_size)
#define EXT2_FRAGS_PER_BLOCK(s)	( (s)->s_frags_per_block)


/*
 * Macro-instructions used to manage group descriptors
 */
#define EXT2_BLOCKS_PER_GROUP(s)	( (s)->s_blocks_per_group)
#define EXT2_DESC_PER_BLOCK(s)		( (s)->s_desc_per_block)
#define EXT2_INODES_PER_GROUP(s)	( (s)->s_inodes_per_group)


/*
 * Structure of the super block on the disk
 */

typedef struct
{
	uint32  	s_inodes_count;		/* Inodes count */
	uint32  	s_blocks_count;		/* Blocks count */
	uint32  	s_r_blocks_count;	/* Reserved blocks count */
	uint32  	s_free_blocks_count;/* Free blocks count */
	uint32  	s_free_inodes_count;/* Free inodes count */
	uint32  	s_first_data_block;	/* First Data Block */
	uint32  	s_log_block_size;	/* Block size */
	int32       s_log_frag_size;	/* Fragment size */
	uint32  	s_blocks_per_group;	/* # Blocks per group */
	uint32  	s_frags_per_group;	/* # Fragments per group */
	uint32  	s_inodes_per_group;	/* # Inodes per group */
	uint32  	s_mtime;			/* Mount time */
	uint32  	s_wtime;			/* Write time */
	uint16 		s_mnt_count;		/* Mount count */
	int16       s_max_mnt_count;	/* Maximal mount count */
	uint16 		s_magic;			/* Magic signature */
	uint16 		s_state;			/* File system state */
	uint16 		s_errors;			/* Behaviour when detecting errors */
	uint16 		s_pad;
	uint32  	s_lastcheck;		/* time of last check */
	uint32  	s_checkinterval;	/* max. time between checks */
	uint32  	s_creator_os;		/* OS */
	uint32  	s_rev_level;		/* Revision level */
	uint32  	s_reserved[236];	/* Padding to the end of the block */
} ext2_super_block;


/*
 * Special inodes numbers
 */
#define	EXT2_BAD_INO		 1l	/* Bad blocks inode */
#define EXT2_ROOT_INO		 2l	/* Root inode */
#define EXT2_ACL_IDX_INO	 3l	/* ACL inode */
#define EXT2_ACL_DATA_INO	 4l	/* ACL inode */
#define EXT2_BOOT_LOADER_INO 5l	/* Boot loader inode */
#define EXT2_UNDEL_DIR_INO	 6l	/* Undelete directory inode */
#define EXT2_FIRST_INO		11l	/* First non reserved inode */


/*
 * Structure of an inode on the disk
 * Please do not remove the union by removing the struct definiftions
 * for other OS's as it affects the size of struct ext2_inode
 */

typedef struct 
{
	uint16 		i_mode;				/* File mode */
	uint16 		i_uid;				/* Owner Uid */
	uint32  	i_size;				/* Size in bytes */
	uint32  	i_atime;			/* Access time */
	uint32  	i_ctime;			/* Creation time */
	uint32  	i_mtime;			/* Modification time */
	uint32  	i_dtime;			/* Deletion Time */
	uint16 		i_gid;				/* Group Id */
	uint16 		i_links_count;		/* Links count */
	uint32  	i_blocks;			/* Blocks count */
	uint32  	i_flags;			/* File flags */
	union
	{
		struct
		{
			uint32  l_i_reserved1;
		} linux1;
		struct
		{
			uint32  h_i_translator;
		} hurd1;
		struct
		{
			uint32  m_i_reserved1;
		} masix1;
	} osd1;							/* OS dependent 1 */

	uint32  	i_block[EXT2_N_BLOCKS];	/* Pointers to blocks */
	uint32  	i_version;				/* File version (for NFS) */
	uint32  	i_file_acl;				/* File ACL */
	uint32  	i_dir_acl;				/* Directory ACL */
	uint32  	i_faddr;				/* Fragment address */
	union
	{
		struct
		{
			uint8  l_i_frag;		/* Fragment number */
			uint8  l_i_fsize;		/* Fragment size */
			uint16 i_pad1;
			uint32 l_i_reserved2[2];
		} linux2;
		struct
		{
			uint8  h_i_frag;		/* Fragment number */
			uint8  h_i_fsize;		/* Fragment size */
			uint16 h_i_mode_high;
			uint16 h_i_uid_high;
			uint16 h_i_gid_high;
			uint32 h_i_author;
		} hurd2;
		struct
		{
			uint8  m_i_frag;		/* Fragment number */
			uint8  m_i_fsize;		/* Fragment size */
			uint16 m_pad1;
			uint32 m_i_reserved2[2];
		} masix2;
	} osd2;							/* OS dependent 2 */
} ext2_inode;

#define i_reserved1	osd1.linux1.l_i_reserved1
#define i_frag		osd2.linux2.l_i_frag
#define i_fsize		osd2.linux2.l_i_fsize
#define i_reserved2	osd2.linux2.l_i_reserved2


/*
 * Structure of a directory entry
 */
#define EXT2_NAME_LEN 255

typedef struct 
{
	uint32  inode;			/* Inode number */
	uint16  rec_len;			/* Directory entry length */
	uint16  name_len;		/* Name length */
	char    name[EXT2_NAME_LEN];	/* File name */
} ext2_dir_entry ;



typedef struct
{
	uint32  bg_block_bitmap;		/* Blocks bitmap block */
	uint32  bg_inode_bitmap;		/* Inodes bitmap block */
	uint32  bg_inode_table;		/* Inodes table block */
	uint16  bg_free_blocks_count;	/* Free blocks count */
	uint16  bg_free_inodes_count;	/* Free inodes count */
	uint16  bg_used_dirs_count;	/* Directories count */
	uint16  bg_pad;
	uint32  bg_reserved[3];
} ext2_group_desc;



/*
 * ACL structures
 */
typedef struct
{
	uint32 aclh_size;
	uint32 aclh_file_count;
	uint32 aclh_acle_count;
	uint32 aclh_first_acle;
} ext2_acl_header;	/* Header of Access Control Lists */

typedef struct
{
	uint32  acle_size;
	uint16 acle_perms;	/* Access permissions */
	uint16 acle_type;	/* Type of entry */
	uint16 acle_tag;	/* User or group identity */
	uint16 acle_pad1;
	uint32  acle_next;	/* Pointer on next entry for the */
					/* same inode or on next free entry */
} ext2_acl_entry;	/* Access Control List Entry */

/*
 * The following is not needed anymore since the descriptors buffer
 * heads are now dynamically allocated
 */
/* #define EXT2_MAX_GROUP_DESC	8 */

#define EXT2_MAX_GROUP_LOADED	8

/*
 * second extended-fs super-block data in memory
 */
typedef struct {
	uint32 s_frag_size;	/* Size of a fragment in bytes */
	uint32 s_frags_per_block;/* Number of fragments per block */
	uint32 s_inodes_per_block;/* Number of inodes per block */
	uint32 s_frags_per_group;/* Number of fragments in a group */
	uint32 s_blocks_per_group;/* Number of blocks in a group */
	uint32 s_inodes_per_group;/* Number of inodes in a group */
	uint32 s_itb_per_group;	/* Number of inode table blocks per group */
	uint32 s_db_per_group;	/* Number of descriptor blocks per group */
	uint32 s_desc_per_block;	/* Number of group descriptors per block */
	uint32 s_groups_count;	/* Number of groups in the fs */

//	struct buffer_head * s_sbh;	/* Buffer containing the super block */
//	struct ext2_super_block * s_es;	/* Pointer to the super block in the buffer */
	ext2_group_desc ** s_group_desc;

	uint16 s_loaded_inode_bitmaps;
	uint16 s_loaded_block_bitmaps;
	uint32 s_inode_bitmap_number[EXT2_MAX_GROUP_LOADED];

//	struct buffer_head * s_inode_bitmap[EXT2_MAX_GROUP_LOADED];
	char * s_inode_bitmap[EXT2_MAX_GROUP_LOADED];

	uint32 s_block_bitmap_number[EXT2_MAX_GROUP_LOADED];

//	struct buffer_head * s_block_bitmap[EXT2_MAX_GROUP_LOADED];
	char * s_block_bitmap[EXT2_MAX_GROUP_LOADED];

//	int32  s_rename_lock;

//	struct wait_queue * s_rename_wait;


//	uint32  s_mount_opt;
//	uint16 s_mount_state;
} ext2_sb_info;




uint32 			ext2_bmap(ldev_t *ldev,unsigned l_unit, ext2_inode far * inode, int32 block);
cache_entry_t* 	ext2_readinode(ldev_t *ldev,unsigned l_unit, uint32 i_ino);
void			ext2_releaseinode(ldev_t *ldev, cache_entry_t *entry);
cache_entry_t* 	ext2_readblock(ldev_t *ldev,unsigned l_unit, uint32 lblock);
void			ext2_releaseblock(ldev_t *ldev, cache_entry_t *entry);
int				ext2_init(ldev_t *ldev,unsigned ino_pages, unsigned block_pages);
void			ext2_shutdown(ldev_t *ldev);
void			ext2_savecontext(ldev_t *ldev);
void			ext2_restorecontext(ldev_t *ldev);
long			ext2_blockscount(ldev_t *ldev);
long			ext2_freeblockscount(ldev_t *ldev);



typedef struct
{
	ext2_super_block 	*ext2_sb;
	ext2_sb_info 		*ext2_sbinfo;
} ext2_fs_info_t;

#endif		// __EXT2_FS_H
