#pragma check_stack(off)
#include <stdlib.h>
#include <dos.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <bios.h>
#include <time.h>

#include "systypes.h"
#include "vsprintf.h"
#include "tsr.h"
#include "ems.h"
#include "diskdev.h"
#include "debug.h"
#include "vfs.h"
#include "redir.h"

#define FILE_NOT_FOUND		2
#define PATH_NOT_FOUND		3
#define ACCESS_DENIED 		5
#define NO_MORE_FILES	   18

char *VolumeString="HD           ";
char *TsrName="MOUNTE2";
char *VersionString="1.0";

typedef struct
{
	uint8 far *		our_int2f_handler;
	uint8 far *		prev_int2f_handler;
	uint8			our_drive;
} user_data_t;

user_data_t		UserData;
int				emskeep;


/* ****************************************************
   Constants and Macros
   **************************************************** */

#define		TRUE				1
#define		FALSE				0

#define		STACK_SIZE			1024
#define		FCARRY				0x0001
#define		MAX_FXN_NO			0x2E


char *usage_string = "Usage:   mounte2 [hda3] d:\r\n\r\n"
	"   hda3: linux partition on disk, first found if ommited \r\n"
	"   d:  : drive on which the ext2fs should be mapped\r\n"
	"Options:\r\n"
	"   DC  : Debug to console (this is the default)\r\n"
	"   DEx : Debug to ems, use x pages (defaults to 1)\r\n"
	"   K   : Keep emspages after unloading\r\n"
	"   Lx  : Debug level x (defaults to 0)\r\n"
	"   M   : Maximum cache (512kb - 1mb)\r\n"
	"   Nx  : # of char's to make a file unique (defaults to 1)\r\n"
	"   SP  : Show partition tables\r\n"
	"   SF  : Show list of debug flags\r\n"
	"   U   : Unload latest version\r\n"
	"\r\n";



INTVECT	prev_int2f_handler;

char far *	cds_curpath;		/* ptr to current path in CDS */
uint8 far * sda_ptr;			/* ptr to SDA */
char far *	sda_filename;		/* ptr to 1st filename area in SDA */
char far *	sda_fcbname;		/* ptr to 1st FCB-style name in SDA */
uint8 far *	sda_srch_attr;		/* ptr to search attribute in SDA */
SRCHREC_PTR	sda_srchrec;		/* ptr to 1st Search Data Block in SDA */
DIRREC_PTR	sda_dirrec;			/* ptr to 1st found dir entry area in SDA */

/* Other global data items */
ALL_REGS	GlobReg;			/* Global save area for all caller's regs */
uint8		our_drive_no;		/* A: is 1, B: is 2, etc. */
char		our_drive_str[3] = " :";	/* Our drive letter string */
char far *	cds_path_root = "Linux :\\";	/* Root string for CDS */
uint16		cds_root_size;		/* Size of our CDS root string */
LOLREC_PTR	lolptr;				/* pointer to List Of Lists */
uint16		dos_ss;				/* DOS's saved SS at entry */
uint16		dos_sp;				/* DOS's saved SP at entry */
uint16		our_sp;				/* SP to switch to on entry */
uint16		save_sp;			/* SP saved across internal DOS calls */
int			filename_is_char_device;	/* generate_fcbname found character device name */
char		our_stack[STACK_SIZE];	/* our internal stack */
uint16 far*	stack_param_ptr;	/* ptr to word at top of stack on entry */
int			curr_fxn;			/* Record of function in progress */

uint8		fxnmap[] = 
{
	_inquiry,				/* 0x00h */
	_rd,					/* 0x01h */
	_unsupported,			/* 0x02h */
	_md,					/* 0x03h */
	_shutdown,				/* 0x04h */
	_cd,					/* 0x05h */
	_closefile,				/* 0x06h */
	_commitfile,			/* 0x07h */
	_readfile,				/* 0x08h */
	_writefile,				/* 0x09h */
	_lockfile,				/* 0x0Ah */
	_unlockfile,			/* 0x0Bh */
	_diskspace,				/* 0x0Ch */
	_unsupported,			/* 0x0Dh */
	_setfileattr,			/* 0x0Eh */
	_getfileattr,			/* 0x0Fh */
	_unsupported,			/* 0x10h */
	_renamefile,			/* 0x11h */
	_unsupported,			/* 0x12h */
	_deletefile,			/* 0x13h */
	_unsupported,			/* 0x14h */
	_unsupported,			/* 0x15h */
	_openfile,				/* 0x16h */
	_createfile,			/* 0x17h */
	_unsupported,			/* 0x18h */
	_unsupported,			/* 0x19h */
	_unsupported,			/* 0x1Ah */
	_findfirst,				/* 0x1Bh */
	_findnext,				/* 0x1Ch */
	_unsupported,			/* 0x1Dh */
	_unsupported,			/* 0x1Eh */
	_unsupported,			/* 0x1Fh */
	_unsupported,			/* 0x20h */
	_seekfromend,			/* 0x21h */
	_unsupported,			/* 0x22h */
	_unsupported,			/* 0x23h */
	_unsupported,			/* 0x24h */
	_unsupported,			/* 0x25h */
	_unsupported,			/* 0x26h */
	_unsupported,			/* 0x27h */
	_unsupported,			/* 0x28h */
	_unsupported,			/* 0x29h */
	_unsupported,			/* 0x2Ah */
	_unsupported,			/* 0x2Bh */
	_unsupported,			/* 0x2Ch */
	_unknown_fxn_2D,		/* 0x2Dh */
	_specialopenfile		/* 0x2Eh */
};



/***************************************************
 * Internal Dos Calls
 ***************************************************/
void set_sft_owner(SFTREC_PTR sft)
	{
	_asm {
		push es
        push di
        les di, sft
        mov save_sp, sp     /* Save current stack pointer. */
        cli
        mov ss, dos_ss      /* Establish DOS's stack, current */
        mov sp, dos_sp      /* when we got called. */
        sti
        mov ax, 0x120c      /* Claim file as ours. */
        push ds             /* It needs DS to be DOS's DS, for DOS 5.0 */
        mov ds, GlobReg.ds
        int 0x2F
        pop bx              /* Restore DS */
        mov ds, bx          /* Restore SS (same as DS) */
        cli
        mov ss, bx
        mov sp, save_sp     /* and stack pointer (which we saved). */
        sti
        pop di
        pop es
    }
}

// Does sda_fcbname point to a device name?
int is_a_character_device(uint16 dos_ds)
{
	_asm {
		mov ax, 0x1223		/* Search for device name. */
		push ds				/* It needs DS to be DOS's DS, for DOS 5.0 */
		mov ds, dos_ds
		int 0x2F
		pop ds				/* Restore DS */
		jnc is_indeed
	}
	return FALSE;
is_indeed:
	return TRUE;
	
}



/*
 * Some frequently used service routines
 */

/* Fail, print message, exit to DOS */
void failprog(char * msg)
{
	tsr_print_string((uint8 far *) msg, TRUE);
	exit(1);
}

void get_fcbname_from_path(char far* path, char far* fcbname)
{
	int i;

	_fmemset(fcbname, ' ', 11);
	for (i = 0; *path; path++)
		if (*path == '.')
			i = 8;
		else
			fcbname[i++] = *path;
}


/*
 * Fail the current redirector call with the supplied error number, i.e.
 * set the carry flag in the returned flags, and set ax=error code
 */

void fail(uint16 err)
{
	GlobReg.flags=(GlobReg.flags | FCARRY);
	GlobReg.ax = err;
}

void succeed(void)
{
	GlobReg.flags=(GlobReg.flags & ~FCARRY);
	GlobReg.ax = 0;
}


/*
 * Deal with differences in DOS version once, and set up a set
 * of absolute pointers
 */

void set_up_pointers(void)
{
	if (_osmajor==3)
	{
		sda_fcbname = ((V3_SDA_PTR) sda_ptr)->fcb_name;
		sda_filename = ((V3_SDA_PTR) sda_ptr)->file_name + cds_root_size - 1;
		//sda_fcbname_2 = ((V3_SDA_PTR) sda_ptr)->fcb_name_2;
		//sda_filename_2 = ((V3_SDA_PTR) sda_ptr)->file_name_2 + cds_root_size - 1;
		sda_srchrec = &((V3_SDA_PTR) sda_ptr)->srchrec;
		sda_dirrec = &((V3_SDA_PTR) sda_ptr)->dirrec;
		//sda_srchrec_2 = &((V3_SDA_PTR) sda_ptr)->rename_srchrec;
		//sda_dirrec_2 = &((V3_SDA_PTR) sda_ptr)->rename_dirrec;
		sda_srch_attr = &((V3_SDA_PTR) sda_ptr)->srch_attr;
	}
	else
	{
		sda_fcbname = ((V4_SDA_PTR) sda_ptr)->fcb_name;
		sda_filename = ((V4_SDA_PTR) sda_ptr)->file_name + cds_root_size - 1;
		//sda_fcbname_2 = ((V4_SDA_PTR) sda_ptr)->fcb_name_2;
		//sda_filename_2 = ((V4_SDA_PTR) sda_ptr)->file_name_2 + cds_root_size - 1;
		sda_srchrec = &((V4_SDA_PTR) sda_ptr)->srchrec;
		sda_dirrec = &((V4_SDA_PTR) sda_ptr)->dirrec;
		//sda_srchrec_2 = &((V4_SDA_PTR) sda_ptr)->rename_srchrec;
		//sda_dirrec_2 = &((V4_SDA_PTR) sda_ptr)->rename_dirrec;
		sda_srch_attr = &((V4_SDA_PTR) sda_ptr)->srch_attr;
	}
}



/*
 * This function should not be necessary. DOS usually generates an FCB
 * style name in the appropriate SDA area. However, in the case of
 * user input such as 'CD ..' or 'DIR ..' it leaves the fcb area all
 * spaces. So this function needs to be called every time. Its other
 * feature is that it uses an internal DOS call to determine whether
 * the filename is a DOS character device. We will 'Access deny' any
 * use of a char device explicitly directed to our drive
 */

void generate_fcbname(uint16 dos_ds)
{
	get_fcbname_from_path((char far*) (_fstrrchr(sda_filename, '\\') + 1),
		sda_fcbname);

	filename_is_char_device = is_a_character_device(dos_ds);
}


/*
 * Does the supplied string contain a wildcard '?'
 */
int contains_wildcards(char far* path)
{
	int i;
	
	for (i = 0; i < 11; i++)
		if (path[i] == '?')
			return TRUE;
	return FALSE;
}

/*
 * Get DOS version, address of Swappable DOS Area, and address of
 * DOS List of lists. We only run on versions of DOS >= 3.10, so
 * fail otherwise
 */

void get_dos_vars(void)
{
	uint16 segmnt;
	uint16 ofset;

	if ((_osmajor < 3) || ((_osmajor == 3) && (_osminor < 10)))
		failprog("Unsupported DOS Version");

	_asm {
		push ds
		push es
		mov ax, 5d06h;	/* Get SDA pointer */
		int 21h;
		mov segmnt, ds
		mov ofset, si
		pop es
		pop ds
	}
	sda_ptr = MK_FP(segmnt, ofset);

	_asm {
		push ds
		push es
		mov ax, 5200h;	/* Get Lol pointer */
		int 21h;
		mov segmnt, es
		mov ofset, bx
		pop es
		pop ds
		}
	lolptr = (LOLREC_PTR) MK_FP(segmnt, ofset);
}


int is_call_for_us(uint16 es, uint16 di, uint16 ds)
{
	uint8 far * p;
	int ret = 0xFF;

	filename_is_char_device = 0;
	
	/* 
	 * Note that the first 'if' checks for the bottom 6 bits
	 * of the device information word in the SFT. Values > last drive
	 * relate to files not associated with drives, such as LAN Manager
	 * named pipes (Thanks to Dave Markun).
	 */

	if ((curr_fxn >= _closefile && curr_fxn <= _unlockfile)
		||	(curr_fxn == _seekfromend)
		||  (curr_fxn == _unknown_fxn_2D) )
	{
		ret = ((((SFTREC_PTR) MK_FP(es,di))->dev_info_word & 0x3F)
			== our_drive_no);
	}
	else
	{
		if (curr_fxn ==_inquiry || curr_fxn==_shutdown)	// 2F/1100 -- succeed automatically
			  ret = TRUE;
		else
		{
			if (curr_fxn == _findnext)		// Find Next
			{
				SRCHREC_PTR psrchrec;	// check search record in SDA
				if (_osmajor==3)
					psrchrec=&(((V3_SDA_PTR) sda_ptr)->srchrec);
				else
					psrchrec=&(((V4_SDA_PTR) sda_ptr)->srchrec);
				return (int) ((psrchrec->drive_no & (uint8) 0x40) &&
					((psrchrec->drive_no & 0x1F) == our_drive_no));
			}
			if (_osmajor==3)
				p=((V3_SDA_PTR) sda_ptr)->cdsptr;	// check CDS
			else
				p=((V4_SDA_PTR) sda_ptr)->cdsptr;

			if (_fmemcmp(cds_path_root, p, cds_root_size) == 0)
			{
				// If a path is present, does it refer to a character device
				if (curr_fxn != _diskspace)
					generate_fcbname(ds);
				return TRUE;
			}
			else
				return FALSE;
		}
	}
	return ret;
}


/*
 * Check to see that we are allowed to install
 */

void is_ok_to_load(void)
{
	_asm {
		mov ax, 1100h;
		int 2fh;
		cmp ax,1
		jne go_forward;
	}
	failprog("Not OK to install a redirector...");

go_forward:
	return;
}


/*
 * This is where we do the initializations of the DOS structures
 * that we need in order to fit the mould
 */

void set_up_cds(void)
{
	V3_CDS_PTR our_cds_ptr;

	our_cds_ptr = lolptr->cds_ptr;
	if (_osmajor == 3) 
		our_cds_ptr = our_cds_ptr + our_drive_no;
	else
	{
		V4_CDS_PTR t = (V4_CDS_PTR) our_cds_ptr;
		t = t + our_drive_no;
		our_cds_ptr = (V3_CDS_PTR) t;
	}

	if (our_drive_no >= lolptr->last_drive)
		failprog("Drive letter higher than last drive.");

	/*
	 * Check that this drive letter is currently invalid (not in use already)
	 * 0xc000 tests both physical and network bits at same time
	 */

	if ((our_cds_ptr->flags & 0xc000) != 0)
		failprog("Drive already assigned...");
		
	/*
	 *  Set Network+Physical+NotRealNetworkDrive bits on, and
	 * establish our 'root'
	 */
	our_cds_ptr->flags |= 0xc080;
	cds_root_size = _fstrlen(cds_path_root);
	_fstrcpy(our_cds_ptr->current_path, cds_path_root);
	our_cds_ptr->current_path[_fstrlen(our_cds_ptr->current_path) - 3] =
		(char) ('A'+ our_drive_no);
	_fstrcpy(cds_path_root, our_cds_ptr->current_path);
	cds_curpath = our_cds_ptr->current_path;
	our_cds_ptr->root_ofs = _fstrlen(our_cds_ptr->current_path) - 1;
	cds_curpath += our_cds_ptr->root_ofs;
}



/*
 * All the redirector routines, remember, for the moment we're
 * a read-only device
 */


/*
 * Respond that it is OK to load another redirector
 */

void redir_inquiry(void)				// subfunc 01
{
	GlobReg.ax = 0x00FF;
}

void redir_readonlydevice(void)		// subfunc 01
{
	fail(ACCESS_DENIED);
}

void redir_findnext(void)				// subfunc 1C
{

	dprintf(D_REDIR,"finding next...\n");
	if (!(vfs_findnext(0, sda_srchrec, sda_dirrec)))
	{
		fail(NO_MORE_FILES);
		return;
	}
	succeed();
}

/*
 * FindFirst - subfunction 1Bh
 */

void redir_findfirst(void)
{

	dprintf(D_REDIR,"findfirst:%fs, mask:%x\n",sda_filename, (unsigned) *sda_srch_attr);
	/*
	 * Handle volumesearch, must be in root
	 */

	if (*sda_srch_attr == ATTR_VOLUME)
	{
		if(sda_filename != _fstrrchr(sda_filename,'\\'))
		{
			fail(PATH_NOT_FOUND);
			return;
		}
		else
		{
			 fail(PATH_NOT_FOUND);	
			_fmemcpy(sda_dirrec->file_name,VolumeString,11);
			sda_dirrec->file_attr=ATTR_VOLUME;
			sda_dirrec->file_size=0;
			return;
		}
	}

	if (!vfs_initsearch(0, sda_filename))
	{
		fail(PATH_NOT_FOUND);
		return;
	}

	_fmemcpy(&sda_srchrec->srch_mask, sda_fcbname, 11);

	sda_srchrec->dir_entry_no = 0;
	sda_srchrec->attr_mask = *sda_srch_attr;
	sda_srchrec->drive_no = (uint8) (our_drive_no | 0xC0);


	redir_findnext();

	/*
	 * No need to check GlobReg.flags & FCARRY; if ax is 18,
	 * FCARRY must have been set.
	 */

	if (GlobReg.ax == 18)
		GlobReg.ax = 2;		// make findnext error code suitable to findfirst
}




/*
 * Change Directory - subfunction 05h
 */

void redir_cd(void)
{
	dprintf(D_REDIR,"cd to:%fs\n",sda_filename);
	if ((*sda_filename != '\\') || (*(sda_filename + 1)))
	{
		if (contains_wildcards(sda_fcbname))
		{
			fail(PATH_NOT_FOUND);
			return;
		}

		*sda_srch_attr = 0x10;
		redir_findfirst();
		if (GlobReg.ax || (! (sda_dirrec->file_attr & 0x10)))
		{
			fail(PATH_NOT_FOUND);
			return;
		}
	}
	_fstrcpy(cds_curpath, sda_filename);
}


/*
 * Close File - subfunction 06h
 */

void redir_closefile(void)
{
	SFTREC_PTR p = (SFTREC_PTR) MK_FP(GlobReg.es, GlobReg.di);

	dprintf(D_REDIR,"close:%fs\n",p->file_name); 
	if (p->handle_count)		/* If handle count not 0, decrement it */
		--p->handle_count;
}


/*
 * Read from File - subfunction 08h
 * For version that handles critical errors,
 * see Undocumented DOS, 2nd edition, chapter 8
 */

void redir_readfile(void)
{
	SFTREC_PTR p = (SFTREC_PTR) MK_FP(GlobReg.es, GlobReg.di);

	dprintf(D_REDIR,"readfile:%fs, file_size:%lu, file_pos:%lu count:%u\n",p->file_name,p->file_size,p->file_pos,GlobReg.cx);
	if (p->open_mode & 1)
	{
		fail(ACCESS_DENIED);
		return;
	}
	
	if ((p->file_pos + GlobReg.cx) > p->file_size)
		GlobReg.cx = (uint16) (p->file_size - p->file_pos);

	if (! GlobReg.cx)
		return;

			// V3_SDA_PTR cast is save: current_dta has fixed offset in SDA
	GlobReg.cx=vfs_readfile(0,p, GlobReg.cx, ((V3_SDA_PTR) sda_ptr)->current_dta);
}


/*
 * Get Disk Space - subfunction 0Ch
 */

void redir_diskspace(void)
{
	dprintf(D_REDIR,"diskspace\n");
	GlobReg.ax = vfs_secpercluster(0);
	GlobReg.bx = vfs_clusters(0);
	GlobReg.cx = vfs_sectorsize(0);;
	GlobReg.dx = vfs_freeclusters(0);;
}


/*
 * Get File Attributes - subfunction 0Fh
 */

void redir_getfileattr(void)
{
	dprintf(D_REDIR,"getfileattrib:%fs\n",sda_filename);
	if (contains_wildcards(sda_fcbname))
	{
		fail(FILE_NOT_FOUND);
		return;
	}

	*sda_srch_attr = 0x3f;
	redir_findfirst();
	if (GlobReg.ax)
		return;
	
	GlobReg.ax = (uint16) sda_dirrec->file_attr;
			/* should we set BX:DI --> filesize ?? */
}



/*
 * Open Existing File - subfunction 16h
 */

void redir_openfile(void)
{
	SFTREC_PTR p=(SFTREC_PTR) MK_FP(GlobReg.es, GlobReg.di);

	dprintf(D_REDIR,"open:%fs\n",sda_fcbname);
	
	if (contains_wildcards(sda_fcbname))
	{
		fail(PATH_NOT_FOUND);
		return;
	}
	
	*sda_srch_attr = 0x27;
	redir_findfirst();

/*
 * If we have a FILE_NOT_FOUND, we do not return immediately, but
 * give the VFS a chance to handle 'special files'
 */

	if (GlobReg.ax)
		return;

	vfs_fill_sft(0,p, sda_dirrec, sda_fcbname);

    if (p->open_mode & 0x8000)
			/* File is being opened via FCB */
       	p->open_mode |= 0x00F0;
   	 else
       	p->open_mode &= 0x000F;

	/*
	 * Mark file as being on network drive, unwritten to
	 */

    p->dev_info_word = (uint16) (0x8040 | (uint16) our_drive_no);
    p->dev_drvr_ptr = NULL;
	p->file_pos=0l;

    dprintf(D_REDIR,"fill_sft opening:%11fs ,mode:%x\n", 
			p->file_name,p->open_mode);

	if(p->open_mode & 0x8000)
		set_sft_owner(p);
}

void redir_createfil(void)
{
	dprintf(D_REDIR,"create file:%fs\n",sda_fcbname);
	fail(ACCESS_DENIED);
	return;
}


/* Special Multi-Purpose Open File - subfunction 2Eh */

#define CREATE_IF_NOT_EXIST		0x10
#define OPEN_IF_EXISTS			0x01
#define REPLACE_IF_EXISTS		0x02

void redir_special_openfile(void)
{
	SFTREC_PTR p = (SFTREC_PTR) MK_FP(GlobReg.es,GlobReg.di);
	unsigned open_mode, action;

	open_mode = ((V4_SDA_PTR) sda_ptr)->mode_2E & 0x7f;
	action = ((V4_SDA_PTR) sda_ptr)->action_2E;
	p->open_mode = open_mode;
	dprintf(D_REDIR,"special open:%fs, mode:%u, action:%u\n",sda_fcbname,open_mode,action);

	if (contains_wildcards(sda_fcbname))
	{
		fail(PATH_NOT_FOUND);
		return;
	}

	*sda_srch_attr = 0x3f;
	redir_findfirst();
	if ((GlobReg.flags & FCARRY) && (GlobReg.ax != 2))
		return;


	if (GlobReg.ax)
	{
		/*
		 * File does not exist
		 */

		if (action & 0xF0)
		{
			fail(ACCESS_DENIED);			// attempt to create
			return;
		}
		else
			return;			// we had a fail(FILE_NOT_FIND)
	}
	else
	{
		/*
		 * File exists
		 */

		if ((action &0x0F) != 1)
		{
			fail(ACCESS_DENIED);		// attempt to replace file
			return;
		}
	}

	vfs_fill_sft(0,p, sda_dirrec, sda_fcbname);

	if (p->open_mode & 0x8000)
			/* File is being opened via FCB */
		p->open_mode |= 0x00F0;
	else
		p->open_mode &= 0x000F;

	/*
	 * Mark file as being on network drive, unwritten to
	 */

	p->dev_info_word = (uint16) (0x8040 | (uint16) our_drive_no);
	p->file_pos = 0l;
	p->dev_drvr_ptr = NULL;

	dprintf(D_REDIR,"fill_sft special opening:%11fs ,mode:%x\n",
			p->file_name,p->open_mode);

	if(p->open_mode & 0x8000)
		set_sft_owner(p);

	succeed();
}


/*
 * This function is never called! DOS fiddles with position internally
 */

void redir_seekfromend(void)
{
	int32 seek_amnt;
	SFTREC_PTR p;

	/* But, just in case... */
	seek_amnt = -1L * (((int32) GlobReg.cx << 16) + GlobReg.dx);
	p = (SFTREC_PTR) MK_FP(GlobReg.es,GlobReg.di);
	if (seek_amnt > p->file_size)
		seek_amnt = p->file_size;
	
	p->file_pos = p->file_size - seek_amnt;
	GlobReg.dx = (uint16) (p->file_pos >> 16);
	GlobReg.ax = (uint16) (p->file_pos & 0xFFFF);
}

void redir_closeall()
{
}

void redir_unknown_fxn_2D()
{
	dprintf(D_REDIR,"Unknown function_2D:%u\n",curr_fxn);
	GlobReg.ax = 2;
	/* Only called in v4.01, this is what MSCDEX returns */
}

void redir_shutdown(void)
{

	dprintf(D_SYSTEM,"Preparing shutdown....\n");
	vfs_shutdown(0);
	// debug_shutdown(); we're cleaned up by the unload process
}

/*
 * A placeholder
 */

void redir_unsupported(void)
{
	dprintf(D_REDIR,"Unsupported function:%u\n",curr_fxn);
	return;
}

PROC dispatch_table[]=
{
	redir_inquiry,				/* 0x00h */
	redir_readonlydevice,		/* 0x01h */
	redir_unsupported,			/* 0x02h */
	redir_readonlydevice,		/* 0x03h */
	redir_shutdown,				/* 0x04h */
	redir_cd,					/* 0x05h */
	redir_closefile,			/* 0x06h */
	redir_readonlydevice,		/* 0x07h */
	redir_readfile,				/* 0x08h */
	redir_readonlydevice,		/* 0x09h */
	redir_unsupported,			/* 0x0Ah */
	redir_unsupported,			/* 0x0Bh */
	redir_diskspace,			/* 0x0Ch */
	redir_unsupported,			/* 0x0Dh */
	redir_readonlydevice,		/* 0x0Eh */
	redir_getfileattr,			/* 0x0Fh */
	redir_unsupported,			/* 0x10h */
	redir_readonlydevice,		/* 0x11h */
	redir_unsupported,			/* 0x12h */
	redir_readonlydevice,		/* 0x13h */
	redir_unsupported,			/* 0x14h */
	redir_unsupported,			/* 0x15h */
	redir_openfile,				/* 0x16h */
	redir_readonlydevice,		/* 0x17h */
	redir_unsupported,			/* 0x18h */
	redir_unsupported,			/* 0x19h */
	redir_unsupported,			/* 0x1Ah */
	redir_findfirst,			/* 0x1Bh */
	redir_findnext,				/* 0x1Ch */
	redir_unsupported,			/* 0x1Dh */
	redir_unsupported,			/* 0x1Eh */
	redir_unsupported,			/* 0x1Fh */
	redir_unsupported,			/* 0x20h */
	redir_seekfromend,			/* 0x21h */
	redir_unsupported,			/* 0x22h */
	redir_unsupported,			/* 0x23h */
	redir_unsupported,			/* 0x24h */
	redir_unsupported,			/* 0x25h */
	redir_unsupported,			/* 0x26h */
	redir_unsupported,			/* 0x27h */
	redir_unsupported,			/* 0x28h */
	redir_unsupported,			/* 0x29h */
	redir_unsupported,			/* 0x2Ah */
	redir_unsupported,			/* 0x2Bh */
	redir_unsupported,			/* 0x2Ch */
	redir_unsupported,			/* 0x2Dh */
	redir_special_openfile		/* 0x0Eh */
	};




/*
 * This is the main entry point for the redirector. It assesses if
 * the call is for our drive, and if so, calls the appropriate routine. On
 * return, it restores the (possibly modified) register values.
 */


void interrupt far redirector(ALL_REGS entry_regs)
{
	static uint16 save_bp;
	static uint16 StkPtr;
	
	_asm STI;

	if (((entry_regs.ax >> 8) != (uint8) 0x11) ||
		((uint8) entry_regs.ax > MAX_FXN_NO))
			goto chain_on;

	curr_fxn = fxnmap[(uint8) entry_regs.ax];

	if ((curr_fxn == _unsupported) ||
		(! is_call_for_us(entry_regs.es, entry_regs.di, entry_regs.ds)))
		goto chain_on;

	/* Set up our copy of the registers */
	GlobReg.ax=entry_regs.ax;
	GlobReg.bx=entry_regs.bx;
	GlobReg.cx=entry_regs.cx;
	GlobReg.dx=entry_regs.dx;
	GlobReg.es=entry_regs.es;
	GlobReg.ds=entry_regs.ds;
	GlobReg.bp=entry_regs.bp;
	GlobReg.sp=entry_regs.sp;
	GlobReg.si=entry_regs.si;
	GlobReg.di=entry_regs.di;
	GlobReg.ip=entry_regs.ip;
	GlobReg.cs=entry_regs.cs;
	GlobReg.flags=entry_regs.flags;
	
/*
 * AAAARRRGGGGHHH, somehow this struct copy causes a hang, donno why
 */
	//GlobReg=entry_regs;

	// Save ss:sp and switch to our internal stack. We also save bp
	// so that we can get at any parameter at the top of the stack
	// (such as the file attribute passed to subfxn 17h).
	_asm cli
	_asm mov dos_ss, ss;
	_asm mov save_bp, bp;
	_asm sti
	//stack_param_ptr = (uint16 far*) MK_FP(dos_ss, save_bp + sizeof(ALL_REGS)); 
/*
 * Make sure the SP is an even address
 */

	StkPtr=(unsigned) our_stack + STACK_SIZE -2;
	if (StkPtr & 1u)
		StkPtr--;
	_asm {
		cli
		mov dos_sp, sp;
		mov ax, ds
		mov ss, ax			// New stack segment is in Data segment.
		//mov sp, offset our_stack + STACK_SIZE -2
		mov sp, StkPtr
		sti
	}

	// Expect success!
	succeed();
	
	// Call the appropriate handling function unless we already know we
	// need to fail
	if (filename_is_char_device)
		fail(ACCESS_DENIED);
	else
	{
		vfs_savecontext();
		dispatch_table[curr_fxn]();
		vfs_restorecontext();
	}

	// Switch the stack back
	_asm {
		cli;
		mov ss, dos_ss;
		mov sp, dos_sp;
		sti;
	}

	// put the possibly changed registers back on the stack, and return
	//entry_regs=GlobReg;
	entry_regs.ax=GlobReg.ax;
	entry_regs.bx=GlobReg.bx;
	entry_regs.cx=GlobReg.cx;
	entry_regs.dx=GlobReg.dx;
	entry_regs.es=GlobReg.es;
	entry_regs.ds=GlobReg.ds;
	entry_regs.bp=GlobReg.bp;
	entry_regs.sp=GlobReg.sp;
	entry_regs.si=GlobReg.si;
	entry_regs.di=GlobReg.di;
	entry_regs.ip=GlobReg.ip;
	entry_regs.cs=GlobReg.cs;
	entry_regs.flags=GlobReg.flags;
	return;

	// If the call wasn't for us, we chain on.
chain_on:
	_chain_intr(prev_int2f_handler);
}


/*
 * Fill in our UserData, upon unloading were are called back with this
 * information
 */

void prepare_for_tsr()
{
	prev_int2f_handler=_dos_getvect(0x2f);
	UserData.prev_int2f_handler=(uint8 far *) prev_int2f_handler;
	UserData.our_int2f_handler=(uint8 far *) redirector;
	UserData.our_drive=our_drive_no;
	_dos_setvect(0x2f,redirector);
}



/*
 * This is called by the tsr_unload routine with info pointing to the
 * the the signature_t of the running trs. If is has been superceded
 * return 0
 */
int unload_redir(signature_t far * sigrec)
{
	INTVECT 		p_vect;
	user_data_t		far *user_data=(user_data_t far *) sigrec->user_data;
	V3_CDS_PTR 		cds_ptr;


/*
 * give the running TSR a change to shutdown properly
 */
	_asm {
		mov	ah,0x11;
		mov al,_shutdown;
		int 2Fh;
	}

	p_vect=_dos_getvect(0x2f);

	/*
	 * Check that a subsequent TSR hasn't taken over Int 2Fh
	 */

	if (user_data->our_int2f_handler != (uint8 far *) p_vect)
	{
		tsr_print_string("Interrupt 2F has been superceded...",1);
		return 0;
	}

	/*
	 * Fix int2f chain
	 */

	p_vect=(INTVECT) user_data->prev_int2f_handler;
	_dos_setvect(0x2f,p_vect);



	our_drive_no=user_data->our_drive;

	cds_ptr=lolptr->cds_ptr;
	if (_osmajor == 3)
		cds_ptr+=our_drive_no;
	else
	{
		V4_CDS_PTR t=(V4_CDS_PTR) cds_ptr;
		t+=our_drive_no;
		cds_ptr=(V3_CDS_PTR) t;
	}

	/*
	 * switch off the Network and Physical bits for the drive,
	 * rendering it invalid.
	 */

	cds_ptr->flags=cds_ptr->flags & 0x3fff;


	our_drive_str[0]=(char) (our_drive_no + 'A');
	tsr_print_string(our_drive_str,FALSE);
	tsr_print_string(" is now invalid.",TRUE);
	if (!emskeep)
		debug_shutdown();
	return 1;
}






/* --------------------------------------------------------------------*/

int _cdecl main(uint16 argc , char **argv)
{
	int disk,partition,dout,dlevel,dpages,showpart,showdlist,maxcache,unload,unique;

	dlevel=showpart=showdlist=0,maxcache=emskeep=unload=0;
	unique=dpages=1;
	dout=DOUT_CONSOLE;
	disk=partition=-1;

	for (argv++; *argv; argv++)
	{
		switch (**argv)
		{
			case '-':
			case '/':
				(*argv)++;
				switch (toupper(**argv)) 
				{
					case 'U':
						unload=1;
						break;
					case 'N':
						unique=simple_atoi((*argv)+1);
						break;
					case 'D':
						(*argv)++;
						switch(toupper(**argv))
						{
							case 'C':
								dout=DOUT_CONSOLE;
								break;
							case 'E':
								dout=DOUT_EMS;
								dpages=simple_atoi((*argv)+1);
								if (!dpages)
									dpages++;		// at least 1 page
								break;
							default:
								dout=DOUT_CONSOLE;
								dlevel=0;
								break;
						}
						break;
					case 'L':
						dlevel=simple_atoi((*argv)+1);
						break;
					case 'M':
						maxcache=1;
						break;
					case 'K':
						emskeep=1;
						break;
					case 'S':
						(*argv)++;
						switch(toupper(**argv))
						{
							case 'P':
								showpart=1;
								break;
							case 'F':
								showdlist=1;
								break;
							default:
								tsr_print_string("Unrecognized parameter.", TRUE);
								return 0;
						}
						break;
					default:
						tsr_print_string("Unrecognized parameter.", TRUE);
						return 0;
				}
				break;
			case 'h':
			case 'H':
				if (*((*argv)+1)==':')
				{
					our_drive_str[0]='H';
					break;
				}

				disk=toupper(*((*argv)+2))-'A';
				partition=*((*argv)+3)-'0';
				if (disk<0 || disk >MAX_HD ||
					partition<0 || partition>MAX_LDEV)
				{
					tsr_print_string("What partition?\r\nUse option SP for partition tables\r\n", TRUE);
					return 0;
				}
				break;
			default:
				our_drive_str[0] = **argv;
		}
	}

	if (unload)
	{
		get_dos_vars();
		tsr_unload_latest(TsrName,unload_redir);
		return 0;
	}



	if (!emsPresent())
	{
		tsr_print_string("Ems not present",TRUE);
		return 0;
	}
	tsr_print_string("MOUNTE2 v.",0);
	tsr_print_string(VersionString,1);

	if (showpart)
	{
		debug_init(D_PARTITION,DOUT_CONSOLE,0);
		pdev_readpartitions();
		return 0;
	}
	if (showdlist)
	{
		debug_init(0,DOUT_CONSOLE,0);
		debug_showlist();
		dprintf(D_ALWAYS,"\nUse binary | to combine debuglevels\n");
		return 0;
	}


		// check that it's a valid drive letter
	if (our_drive_str[0] == ' ')
	{
		tsr_print_string(usage_string, TRUE);
		return 0;
	}

	our_drive_str[0] &= ~0x20;
	our_drive_no = (uint8) (our_drive_str[0] - 'A');
	if (our_drive_no > 25)
	{
		tsr_print_string(usage_string, TRUE);
		return 0;
	}

/*
 * Let's first check that we are 'allowed' to load,
 * and the desired drive is not in use
 */
	is_ok_to_load();
	get_dos_vars();
	set_up_cds();
	set_up_pointers();

/*
 * Prepare debugging and read partitions tables
 */


	debug_init(dlevel, dout, dpages);
	dprintf(D_EMS,"Ems v%s: tot pages:%u, free: %u\n",emsGetVersion(),emsNumTotPages(),emsNumFreePages());
	pdev_readpartitions();

	if (!(vfs_setuppartition(&disk,&partition,maxcache,unique)))
		return 0;
	
	VolumeString[2]=disk+'A';
	if (partition > 10)
	{
		VolumeString[3]='1';
		VolumeString[4]=(partition % 10) +'0';
	}
	else
		VolumeString[3]=partition +'0';


	dout=debug_setout(DOUT_CONSOLE);
	dprintf(D_ALWAYS,"hd%c%c installed as %s\n",disk+'a',partition+'0', our_drive_str);


/*
 * Let's hit it Stan!
 */

	prepare_for_tsr();						// sets UserData and chain int2f
	debug_setout(dout);
	tsr_go_resident(TsrName,&UserData);

	return 0;
}




