#pragma check_stack(off)
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef __BORLANDC__
#include <alloc.h>
#else
#include <malloc.h>
#endif

#include "unixstat.h"
#include "debug.h"
#include "ext2_fs.h"
#include "ems.h"
#include "emscache.h"
#include "diskdev.h"
#include "redir.h"
#include "vfs.h"


/*
 * This module is supposed to be a sort of Virtual File System.
 * It hides the actual file system from the redirector, in our case
 * it's Linux' ext2fs.
 *
 * The 'only' routine that shouldn't be here is ext2_entrylookup.
 *
 * PVS: this interface should definitly get cleaned up. It must provide
 * access to multiple file systems and must support different types
 */

typedef struct
{
	char			srch_localname[128];
	char			srch_filename[11];
	unsigned long   srch_filesize;
	unsigned long   srch_filetime;
	unsigned long   srch_dir_ino;
	unsigned long   srch_entry_ino;
	long 			srch_cur_block;
	long 			srch_block_pos;
	unsigned        srch_offset;
	unsigned		srch_i_mode;
	int				n_phantom;
} srch_block_t;

srch_block_t		ffirst_search;
srch_block_t		origfiles_search;

char				*phantom_name="$FILES  ORG";

ldev_t 				*Our_ldev;

#define SF_CHARDEV 	1
#define SF_BLOCKDEV	2
#define SF_SYMLINK	3
#define SF_SOCKET	4
#define SF_PIPE		5
#define SF_PHANTOM	6

#define PHANTOM_FRONTFORMAT 26

static char			*special_files[]=
					{
						"",
						"This is a character device.",
						"This is a block device.",
						"This is a symbolic link.",
						"This is a socket.",
						"This is a named pipe.",
					};

#define ENTRY_DIR   1
#define ENTRY_FILE  2
#define ENTRY_BOTH	3

#define HIGH_WORD(ul)   ((unsigned) (((ul) & 0xFFFF0000ul ) >> 8))
#define LOW_WORD(ul)    ((unsigned) ((ul) & 0x0000FFFFul))
#define MK_LWORD(u1,u2) ((((unsigned long) (u1)) << 8) | ((unsigned long) (u2)))


int     Digits_To_Make_Unique=1;
char 	bad_chars[]="*?<>\\$+=;";


static void convert_filename(char far *filename, char far *newfilename, int offset)
{
	char far 	*pc;
	char		c;
	int			i,has_to_chop,n_move;

	_fmemset(newfilename,' ',11);
	pc=newfilename;

/*
 * convert hidden files (starting with a dot) to underscore, unless
 * it's the . or .. entry
 */
	if (filename[0]=='.')
	{
		newfilename[0]='.';
		if (filename[1]==0)
			return;
		else
			if (filename[1]=='.' && filename[2]==0)
			{
				newfilename[1]='.';
				return;
			}
			else
				filename[0]='_';
	}
/*
 * copy filename part
 */
	for (i=0 ; i<8 ; i++)
	{
		c=*filename++;
		if (c=='.' || !c)
			break;
		if (strchr(bad_chars,c))
			c='_';
		*pc++=toupper(c);
	}
	has_to_chop= (i==8 && (*filename!='.' && *filename!=0));


/*
 * search dot for extension of the filename
 */

	for(; c && c!='.' ; c=*filename++)
		;

/*
 * copy extension
 */

	pc=newfilename+8;
	for (i=0 ; i<3 && c ; i++)
	{
		c=*filename++;
		if (!c)
			break;
		if (strchr(bad_chars,c) || c=='.')
			c='_';
		*pc++=toupper(c);
	}
/*
 * if either the filename or the extention part is too long,
 * chop it off
 */
	if (has_to_chop || c)
	{
		pc=newfilename+8;
		n_move=i+Digits_To_Make_Unique-1;
		pc+= (n_move>2)? 2 : n_move;
		for (i= 0 ; i<Digits_To_Make_Unique ; i++)
		{
			*pc--= offset % 10 + '0';
			offset/=10;
		}
	}
}

/* 
 * This comes strait from the Linux kernel source. I took it from
 * msdos filesystem. Donno what it does, but it works :-)
 */

/* Linear day numbers of the respective 1sts in non-leap years. */
static long day_n[] = { 0,31,59,90,120,151,181,212,243,273,304,334,0,0,0,0 };
		  /* JanFebMarApr May Jun Jul Aug Sep Oct Nov Dec */

static uint32 date_unix2dos(uint32 unix_date)
{
	long day,year,nl_day,month;
	unsigned long time,date;

	time = (unix_date % 60)/2+(((unix_date/60) % 60) << 5)+
		(((unix_date/3600) % 24) << 11);
	day = unix_date/86400l-3652;
	year = day/365;
	if ((year+3)/4+365*year > day) year--;
	day -= (year+3)/4+365*year;
	if (day == 59 && !(year & 3)) {
		nl_day = day;
		month = 2;
	}
	else {
		nl_day = (year & 3) || day <= 59 ? day : day-1;
		for (month = 0; month < 12; month++)
			if (day_n[month] > nl_day) break;
	}
	date = nl_day-day_n[month-1]+1+(month << 5)+(year << 9);

	return (date<<16) | time;
}

static int match_names(char far *entry_name, char far *masked_name)
{
	char	far *pc1=masked_name,far *pc2=entry_name;
	int 	i;

	for (i=0 ; i<11 ; i++,masked_name++,entry_name++)
	{
		if ( *masked_name != '?' && *masked_name!=*entry_name)
			break;
	}
	dprintf(D_VFS,"match_names:%-11fs,%-11fs, returns:%s\n",pc1,pc2,(i==11)?"True":"False",i);
	return (i==11);
}

static int vfs_isphantomfile(unsigned l_unit, char far *filename)
{
	return match_names(filename,phantom_name);
}

/*
 * Ok, this is the basic part that lookups an entry in an directory (specified
 * by cur_ino). It implements the ffirst/fnext mechanism that is used by DOS. 
 * The goto in the middle of the code is needed to gracefully return without 
 * touching the state we're in.
 */
char *search_type[]={"DIR","FILE","BOTH"};

static unsigned long ext2_entrylookup(unsigned l_unit, srch_block_t *search, char far *entryname,int what)
{
	ext2_dir_entry far	*dirent;
	ext2_inode			far *inode,far *entry_inode;
	cache_entry_t		*e1,*e2,*e3;
	long				block,bsize=BLOCKSIZE(Our_ldev);
	int					is_dir;

	search->srch_entry_ino=0l;
	dprintf(D_VFS,"Looking for %s-entries:%-11fs in dir with ino:%lu\n", search_type[what-1],entryname,search->srch_dir_ino);
	if (!(e1=ext2_readinode(Our_ldev,l_unit,search->srch_dir_ino)))
		return 0l;
	inode=(ext2_inode far*) e1->data;
	dprintf(D_VFS,"ino_size:%lu, ino_blocks:%lu\n", inode->i_size,inode->i_blocks);

	while(search->srch_cur_block*bsize< (long) inode->i_size && !search->srch_entry_ino)
	{
		block=ext2_bmap(Our_ldev,l_unit,inode,search->srch_cur_block);
		if (!(e2=ext2_readblock(Our_ldev,l_unit,block)))
			goto end_entrylookup;
		dirent=(ext2_dir_entry far *) ((char far *) e2->data+search->srch_block_pos);
		while (search->srch_block_pos!=bsize)
		{
			_fstrncpy(search->srch_localname,dirent->name,dirent->name_len);
			search->srch_localname[dirent->name_len]=0;
			convert_filename(search->srch_localname,search->srch_filename,search->srch_offset);

			search->srch_block_pos+=dirent->rec_len;
			search->srch_offset++;
			if(match_names(search->srch_filename,entryname))
			{
				if (!(e3=ext2_readinode(Our_ldev,l_unit,dirent->inode)))
				{
					dirent=(ext2_dir_entry far *) (((char far *) dirent) + dirent->rec_len);
					continue;		//skip this entry
				}
					
				entry_inode=(ext2_inode far*) e3->data;
				is_dir=	S_ISDIR(entry_inode->i_mode) ? 1 : 0;
				search->srch_i_mode=entry_inode->i_mode;
				search->srch_filesize=entry_inode->i_size;
				search->srch_filetime=entry_inode->i_atime;
				dprintf(D_VFS,"Entry-ino:%lu, isdir:%s\n",dirent->inode,is_dir?"Yes":"No");
				ext2_releaseinode(Our_ldev,e3);

				if ( 	what==ENTRY_BOTH || 
						(what==ENTRY_FILE && !is_dir) || 
						(what==ENTRY_DIR && is_dir))
				{
					search->srch_entry_ino=dirent->inode;
					ext2_releaseblock(Our_ldev,e2);
					goto end_entrylookup;
				}
			}
			dirent=(ext2_dir_entry far *) (((char far *) dirent) + dirent->rec_len);
		}
		search->srch_block_pos=0l;
		ext2_releaseblock(Our_ldev,e2);
		search->srch_cur_block++;
	}
end_entrylookup:
	ext2_releaseinode(Our_ldev,e1);
	return search->srch_entry_ino;
}


void vfs_dirlookup(unsigned l_unit, char far *dirname)
{
	char			far *pc,far *pc_ext,far *subdir;
	int				len;
	char			local[11];

	
	ffirst_search.srch_dir_ino=EXT2_ROOT_INO;
	subdir=dirname+1;
	while((pc=_fstrchr(subdir,'\\')) && ffirst_search.srch_dir_ino)
	{
		*pc=0;
		ffirst_search.srch_cur_block=ffirst_search.srch_block_pos=0l;
		ffirst_search.srch_offset=ffirst_search.srch_i_mode=0;
		memset(local,' ',11);
		if ((pc_ext=_fstrchr(subdir,'.')))
		{
			len=pc_ext-subdir;
			_fmemcpy(local,subdir,len);
			_fmemcpy(local+8,subdir+len+1,pc-subdir-len-1);
		}
		else
			_fmemcpy(local,subdir,pc-subdir);
		ffirst_search.srch_dir_ino=ext2_entrylookup(l_unit,&ffirst_search,(char far *)local,ENTRY_DIR);
		subdir=pc+1;
		*pc='\\';
	}
	dprintf(D_VFS,"dirlookup on:%fs,found inode:%lu\n",dirname,ffirst_search.srch_dir_ino);
}

/*
 * Check if disk and partition make up a valid device with the right type
 * if disk==-1, search for the first such a partition. Depending on
 * maxCache we initiale the ext2fs with maximum or minimum cache
 */

int vfs_setuppartition(int *disk, int *partition, int maxcache, int Unique)
{
	int			succes,d,p;

	d=*disk;
	p=*partition;
	Digits_To_Make_Unique=(Unique<1) ? 1: (Unique > 3) ? 3: Unique;
	succes=0;

	if (d==-1)
	{
		dprintf(D_ALWAYS,"Searching for an ext2 file system...\n");
		for (d=0 ; d< MAX_HD ; d++)
			for (p=0 ; p<MAX_LDEV ; p++)
			{
				Our_ldev=&lDevices[d][p];
				if (Our_ldev->type==131)
				{
					dprintf(D_ALWAYS,"Found one on hd%c%c\n",d+'a',p+'0');
					goto got_partition;
				}
			}
		dprintf(D_ALWAYS,"There aren't any!\n");
		return 0;
	}
	else
		Our_ldev=&lDevices[d][p];

got_partition:
	if (Our_ldev->type==131)
	{
		dprintf(D_EXT2,"Found ext2 file system on partition hd%c%c\n",d+'a',p+'0');
		if(maxcache)
			succes=ext2_init(Our_ldev, 2, 48);
		else
			succes=ext2_init(Our_ldev, 1, 1);
		if (!succes)
			dprintf(D_ALWAYS,"Ouch, could not setup ext2 file system!\n");
	}
	else
		dprintf(D_ALWAYS,"Could not find an ext2 file system on hd%c%c\n",d+'a',p+'0');

	*disk=d;
	*partition=p;
	return succes;
}

char old_path[128]="";
unsigned long old_ino;

int vfs_initsearch(unsigned l_unit, char far * path)
{
	char 		far *pc;

	if ((pc=_fstrrchr(path+1,'\\')))
	{
		*pc=0;
		if (_fstrcmp(path,old_path)==0)
		{
			ffirst_search.srch_dir_ino=old_ino;
		}
		else
		{
			*pc='\\';
			vfs_dirlookup(l_unit, path); 		// sets ffirst_search.srch_dir_ino
			*pc=0;
			_fstrcpy(old_path,path);
			old_ino=ffirst_search.srch_dir_ino;
		}
		*pc='\\';
	}
	else
		ffirst_search.srch_dir_ino=EXT2_ROOT_INO;

	ffirst_search.srch_cur_block=ffirst_search.srch_block_pos=0l;
	ffirst_search.srch_offset=ffirst_search.srch_i_mode=0;
	ffirst_search.n_phantom=0;
	dprintf(D_VFS,"initsearch:%fs, found inode:%lu\n",path,ffirst_search.srch_dir_ino);
	return (int) ffirst_search.srch_dir_ino;
}

int vfs_findnext(unsigned l_unit, SRCHREC_PTR srchrec, DIRREC_PTR dirrec)
{
	int	what=(srchrec->attr_mask & ATTR_SUBDIR) ? ENTRY_BOTH : ENTRY_FILE;


	if (ffirst_search.n_phantom)
		return 0;
	
	if (ext2_entrylookup(l_unit,&ffirst_search, srchrec->srch_mask,what))
	{
		dprintf(D_VFS,"ffnext: found entry:%-11s\n",ffirst_search.srch_filename);
		_fmemcpy(dirrec->file_name,ffirst_search.srch_filename,11);
		dirrec->file_attr=S_ISDIR(ffirst_search.srch_i_mode)? ATTR_SUBDIR : ATTR_READONLY;
		dirrec->start_sector=(uint16) ffirst_search.srch_entry_ino;
		dirrec->file_size=S_ISDIR(ffirst_search.srch_i_mode)? 0l: (int32) ffirst_search.srch_filesize;
		dirrec->file_time=date_unix2dos(ffirst_search.srch_filetime);
		srchrec->dir_entry_no=ffirst_search.srch_offset;
		srchrec->dir_sector=(uint16) ffirst_search.srch_dir_ino;
	}
	else
	{
		ffirst_search.n_phantom++;
		if (!match_names(phantom_name,srchrec->srch_mask))
			return 0;

		dprintf(D_VFS,"ffnext: made phantom file:%s\n",phantom_name);
		_fmemcpy(dirrec->file_name,phantom_name,11);
		dirrec->file_attr=ATTR_READONLY;
		dirrec->start_sector=0;
		dirrec->file_size=0;
		dirrec->file_time=date_unix2dos(ffirst_search.srch_filetime);
		srchrec->dir_entry_no=0;
		srchrec->dir_sector=0;
	}

	return 1;
}

/*
 * Fill in the entry in the System File Table. If it's a special file
 * we'll put some different information in this entry (f.e. a pointer
 * to a description string). 
 */

void vfs_fill_sft(unsigned l_unit, SFTREC_PTR p, DIRREC_PTR dirrec, char far *fcbname)
{
    _fmemcpy(p->file_name, fcbname, 11);
    p->file_attr=dirrec->file_attr;
    p->file_time=dirrec->file_time;
    p->rel_sector=HIGH_WORD(ffirst_search.srch_entry_ino);
    p->abs_sector=LOW_WORD(ffirst_search.srch_entry_ino);

	p->dir_sector= 	S_ISCHR(ffirst_search.srch_i_mode) ? SF_CHARDEV :
					S_ISBLK(ffirst_search.srch_i_mode) ? SF_BLOCKDEV :
					S_ISLNK(ffirst_search.srch_i_mode) ? SF_SYMLINK :
					S_ISSOCK(ffirst_search.srch_i_mode) ? SF_SOCKET :
					S_ISFIFO(ffirst_search.srch_i_mode) ? SF_PIPE :
					match_names(fcbname,phantom_name)? SF_PHANTOM : 0;

	if (p->dir_sector == SF_PHANTOM)
	{
		origfiles_search.srch_dir_ino=ffirst_search.srch_dir_ino;
		origfiles_search.srch_cur_block=0;
		origfiles_search.srch_block_pos=0;
		origfiles_search.srch_offset=0;
		p->start_sector=0;
		//p->file_size= 7000;
		p->file_size= (ffirst_search.srch_cur_block+1) * BLOCKSIZE(Our_ldev) +
					PHANTOM_FRONTFORMAT * ffirst_search.srch_offset;
	}
	else
		if (p->dir_sector)
		{
    		p->file_size=strlen(special_files[p->dir_sector]);
    		p->start_sector=(unsigned) special_files[p->dir_sector];
		}
		else
		{
   		 	p->dir_sector=0;
   		 	p->file_size=dirrec->file_size;
   		 	p->start_sector=dirrec->start_sector;
		}
    p->dir_entry_no=(uint8) ffirst_search.srch_offset;
}


/*
 * The $file.org lists for each directory entry, the type, the access
 * list, the converted dos name and the original name. 
 * example:
 * drwxr-xr-x LOST_FOU 2 = lost+found
 */

static char format_buffer[PHANTOM_FRONTFORMAT+1];

static char *format_phantomline(srch_block_t *search)
{
	int		i;
	uint16	mode= (uint16) search->srch_i_mode;

	format_buffer[0] = 	S_ISCHR(search->srch_i_mode) ? 'c' :
				S_ISBLK(search->srch_i_mode) ? 'b' :
				S_ISLNK(search->srch_i_mode) ? 'l' :
				S_ISSOCK(search->srch_i_mode) ? 's' :
				S_ISFIFO(search->srch_i_mode) ? 'p' :
				S_ISDIR(search->srch_i_mode) ? 'd' : ' ';

	for (i=0 ; i<9 ; i++)
	{
		format_buffer[9-i]= (mode & 1) ? "xrw"[i%3] : '-';
		mode>>=1;
	}
	format_buffer[10]=' ';
	_fmemcpy(&format_buffer[11],search->srch_filename,8);
	format_buffer[19]=' ';
	_fmemcpy(&format_buffer[20],search->srch_filename+8,3);
	format_buffer[23]=' ';
	format_buffer[24]='=';
	format_buffer[25]=' ';

	return format_buffer;
}


static unsigned vfs_readphantomfile(SFTREC_PTR p, unsigned size, char far *dst)
{
	char			*src;
	unsigned 		to_go,n,prev_offset;
	unsigned long   ino;

	prev_offset=p->start_sector;
	to_go=size;

	dprintf(D_VFS,"readphantom: read size:%u\n",size);

	if (prev_offset==0)
	{
		ino=ext2_entrylookup(0, &origfiles_search, "???????????",ENTRY_BOTH);
		if (!ino)
		{
			dprintf(D_VFS,"lookup returned inode:%lu\n",ino);
			p->file_size=p->file_pos;
			return 0;
		}
	}
		

	do
	{
		dprintf(D_VFS,"readphantom: %s item: %s\n",prev_offset?"finishing":"adding", origfiles_search.srch_filename);
		if (prev_offset < PHANTOM_FRONTFORMAT )
		{
			src=format_phantomline(&origfiles_search);
			n=PHANTOM_FRONTFORMAT-prev_offset;
			n=(to_go < n) ? to_go : n;
			p->start_sector=n;
			_fmemcpy(dst,src+prev_offset,n);
			to_go-=n;
			prev_offset+=n;
			dst+=n;
		}
		if (to_go)
		{
			n=_fstrlen(origfiles_search.srch_localname+prev_offset-PHANTOM_FRONTFORMAT);
			n=(to_go<n) ? to_go : n;
			p->start_sector+= n;

			_fmemcpy(dst,origfiles_search.srch_localname+prev_offset-PHANTOM_FRONTFORMAT,n);
			to_go-=n;
			dst+=n;
		}
		if (to_go)
		{
			*dst++=10;
			p->start_sector=prev_offset=0;
			to_go--;
		}


	} while (to_go && ext2_entrylookup(0, &origfiles_search, "???????????",ENTRY_BOTH));

	if (to_go)
	{
		p->file_size+=p->file_pos+size-to_go;
	}

	p->file_pos+=size-to_go;
	return size-to_go;
}

static unsigned vfs_readspecialfile(SFTREC_PTR p, unsigned size, char far *dst)
{
	char			*src;
	unsigned 		n,file_pos;
	unsigned long	cur_ino;


	cur_ino=MK_LWORD(p->rel_sector,p->abs_sector);
	file_pos=(unsigned) p->file_pos;
	src = ((char *) p->start_sector) + file_pos;
	n=((unsigned) p->file_size) - file_pos;

	dprintf(D_VFS,"ext2_readspecialfile (%u):%11fs, ino:%ul, fp:%u;, size:%u\n",p->dir_sector,p->file_name,cur_ino,p->file_pos,size);

	n=(size < n) ? size : n;
	_fmemcpy(dst,src,n);
	p->file_pos+=n;
	return n;
}

unsigned vfs_readfile(unsigned l_unit, SFTREC_PTR p, unsigned size, char far *dst)
{
	cache_entry_t		*e1,*e2;
	ext2_inode			far *inode;
	unsigned 			n,to_go;
	unsigned long 		cur_ino, block, lblock,bsize=BLOCKSIZE(Our_ldev),offset;
	char 				far *src;

	if(p->dir_sector==SF_PHANTOM)
		return vfs_readphantomfile(p,size,dst);

	if(p->dir_sector)
		return vfs_readspecialfile(p,size,dst);

	cur_ino=MK_LWORD(p->rel_sector,p->abs_sector);
	lblock=p->file_pos / bsize;
	offset=p->file_pos % bsize;
	to_go=size;

	dprintf(D_VFS,"ext2_readfile:%11fs, ino:%ul, fp:%u;, size:%u\n",p->file_name,cur_ino,p->file_pos,size);
	

	if (!(e1=ext2_readinode(Our_ldev,l_unit,cur_ino)))
	{
		dprintf(D_ALWAYS,"ext_readfile: could not read inode:%lu\n,",cur_ino);
		goto vfs_readfile_end;
	}
	inode=(ext2_inode far*) e1->data;


	for(to_go=size; to_go ; lblock++)
	{
		block=ext2_bmap(Our_ldev,l_unit,inode,lblock);
		if (!(e2=ext2_readblock(Our_ldev,l_unit,block)))
		{
			dprintf(D_ALWAYS,"ext_readfile: readblock %lu failed\n",lblock);
			dprintf(D_ALWAYS,"filepos:%lu, filesz:%lu, size:%u, togo:%u\n",p->file_pos,p->file_size,size,to_go);
			p->file_size=p->file_pos+size-to_go;// make sure next read is EOF!!
			break;
		}
		src=((char far *) e2->data)+offset;
		n=(unsigned) (bsize-offset);
		n=(n>to_go) ? to_go : n;
		_fmemcpy(dst,src,n);
		ext2_releaseblock(Our_ldev,e2);
		offset=0l;
		to_go-=n;
		dst+=n;
	}
	ext2_releaseinode(Our_ldev,e1);
vfs_readfile_end:
	p->file_pos+=size-to_go;
	return size-to_go;
}

void vfs_shutdown(unsigned l_unit)
{
	ext2_shutdown(Our_ldev);
}

void vfs_savecontext()
{
	ext2_savecontext(Our_ldev);
}

void vfs_restorecontext()
{
	ext2_restorecontext(Our_ldev);
}

unsigned vfs_secpercluster(unsigned l_unit)
{
	return Our_ldev->sec_perblock;
}

unsigned vfs_clusters(unsigned l_unit)
{
	return (unsigned) (ext2_blockscount(Our_ldev) / Our_ldev->sec_perblock);
}

unsigned vfs_freeclusters(unsigned l_unit)
{
	return (unsigned) (ext2_freeblockscount(Our_ldev) / Our_ldev->sec_perblock);
}

unsigned vfs_sectorsize(unsigned l_unit)
{
	return BYTES_PERSECTOR;
}


